@@ -5,6 +5,7 @@ import glob
 import itertools
 import warnings
 import numpy as np
+from scipy.interpolate import griddata
 from shapely import wkt
 from shapely.geometry import Point, Polygon, MultiPoint, MultiPolygon, LineString, MultiLineString, LinearRing, mapping, shape
 from shapely.ops import transform, linemerge, unary_union, polygonize
@@ -48,9 +49,10 @@ from lfmaptools.utilities import (_nice_round,
                                   _path_of_file_in_zip,
                                   csv_active_cols,
                                   csv_num_cols)
-from lfmaptools.geotools import LatLon_TM_polygonize
+from lfmaptools.geotools import LatLon_TM_polygonize, LatLon_TM
 from lfmaptools.epsg_defs import webmerc, wgs84
 from lfmaptools.mapping import add_stamen_basemap
+from lfmaptools.interptools import interp_weights, interpolate
 
 _LFoutput_fmt = [
     {'file': 'all', 'column': 0, 'short_name': 'tile', 'long_name': 'tile id', 'symbol': None, 'units': 'dimensionless', 'dtype': int, 'infile': True, },
@@ -129,15 +131,15 @@ class LaharFlowData(object):
                 if "height threshold" in line:
                     self.hthres = float(line.split('=',1)[1])
                 if "nTiles" in line:
-                    self.nTiles = float(line.split('=',1)[1])
+                    self.nTiles = int(line.split('=',1)[1])
                 if "nXtiles" in line:
-                    self.nXtiles = float(line.split('=',1)[1])
+                    self.nXtiles = int(line.split('=',1)[1])
                 if "nYtiles" in line:
-                    self.nYtiles = float(line.split('=',1)[1])
+                    self.nYtiles = int(line.split('=',1)[1])
                 if "nXpertile" in line:
-                    self.nXpertile = float(line.split('=',1)[1])
+                    self.nXpertile = int(line.split('=',1)[1])
                 if "nYpertile" in line:
-                    self.nYpertile = float(line.split('=',1)[1])
+                    self.nYpertile = int(line.split('=',1)[1])
                 if "Xtilesize" in line:
                     self.Xtilesize = float(line.split('=',1)[1])
                 if "Ytilesize" in line:
@@ -211,6 +213,7 @@ class LaharFlowData(object):
             
             self.get_files()
             self.get_txt_files()
+            self.get_nc_files()
             self.get_tar_files()
             self.get_tiff_files()
 
@@ -368,6 +371,97 @@ class LaharFlowData(object):
                     pass
                     
         self.parameters = parameters
+    
+    def _grid_to_physical(self, grid_i, grid_j, tile_i, tile_j, x, y):
+      
+      x = -0.5 * self.xSize + self.dX * ((grid_i - 1.0) * self.nXpertile + (tile_i - 0.5))
+      y = -0.5 * self.ySize + self.dY * ((grid_j - 1.0) * self.nYpertile + (tile_j - 0.5))
+
+      return x, y
+    
+    def _tile_row(self, tile):
+
+      row = (tile - 1) // self.nXtiles + 1
+
+      return int(row)
+    
+    def _tile_col(self, tile):
+      
+      col = (tile - 1)%self.nXtiles + 1
+
+      return int(col)
+
+    def _grid_coords(self, tileID):
+      
+      i = self._tile_col(tileID)
+      j = self._tile_row(tileID)
+
+      return i, j
+    
+    def _tile_id(self, grid_i, grid_j):
+
+        t = grid_i + (grid_j-1)*self.nXtiles
+        return t
+    
+    def _neighbour_tiles(self, tile):
+
+        i, j = self._grid_coords(tile)
+
+        neighbours = []
+
+        for k in [-1,1]:
+            neighbours.append(self._tile_id(i+k, j))
+            neighbours.append(self._tile_id(i, j+k))
+            neighbours.append(self._tile_id(i+k, j+k))
+            neighbours.append(self._tile_id(i+k, j-k))
+        
+        return neighbours
+
+    def _connected_tiles(self, tiles):
+
+        groups = [[tiles[0]]]
+
+        tiles = tiles[1:]
+
+        for t in tiles:
+            neighbours = self._neighbour_tiles(t)
+
+            in_group = False
+            for group in groups:
+                if any(n in group for n in neighbours):
+                    group.append(t)
+                    in_group = True
+                    break
+            
+            if not in_group:
+                groups.append([t])
+    
+        return groups
+    
+    def _tile_row_groups(self, tiles):
+
+        rows = [self._tile_row(t) for t in tiles]
+
+        ur = np.unique(rows)
+
+        row_groups=[]
+        for r in ur:
+            row_groups.append([t for j, t in enumerate(tiles) if rows[j]==r])
+
+        return row_groups
+
+    def _get_latlon_res(self):
+
+        East_step = LatLon_TM(self.clatlong, [self.dX, 0], unit='m')
+        North_step = LatLon_TM(self.clatlong, [0, self.dY], unit='m')
+
+        clat= self.clat
+        clong = 360 + self.clong if self.clong<0 else self.clong
+
+        lat_res = North_step[0] - clat
+        lon_res = East_step[1] - clong
+
+        return (lat_res, lon_res)
 
     def get_XY(self):
         
@@ -387,6 +481,10 @@ class LaharFlowData(object):
         self.tar_files = [f for f in self.files if f.endswith('.tar.gz')]
         self.tar_files.sort()
 
+    def get_nc_files(self):
+        self.nc_files = [f for f in self.files if f.endswith('.nc')]
+        self.nc_files.sort()
+
     def get_txt_files(self):
         self.txt_files = [f for f in self.files if f.endswith('.txt')]
         self.txt_files.sort()
@@ -428,6 +526,8 @@ class LaharFlowData(object):
                 fileIn = '{0:06d}.txt'.format(fileIn)
             elif '{0:06d}.txt.tar.gz'.format(fileIn) in self.files:
                 fileIn = '{0:06d}.txt.tar.gz'.format(fileIn)
+            elif '{0:06d}.nc'.format(fileIn) in self.files:
+                fileIn = '{0:06d}.nc'.format(fileIn)
             else:
                 raise ValueError('Result file {file} is not in directory {dir}'.format(file=fileIn, dir=self.dir))
         else:
@@ -463,7 +563,10 @@ class LaharFlowData(object):
         elif filename in self.tar_files:
             LaharFlowFile = os.path.join(self.dir, filename)
             ExtractPath = self.dir
-        
+        elif filename in self.nc_files:
+            LaharFlowFile = os.path.join(self.dir, filename)
+            ExtractPath = self.dir
+
         if (os.path.split(LaharFlowFile)[1] in self.tar_files) and (os.path.split(LaharFlowFile)[1] not in self.txt_files):
             tar = tarfile.open(LaharFlowFile, mode='r')
             tar.extractall(path=ExtractPath)
@@ -485,14 +588,21 @@ class LaharFlowData(object):
 
         LaharFlowFile = self._extract_file(filename)
         
-        try:
-            cols = csv_active_cols(LaharFlowFile)
-            data = np.genfromtxt(LaharFlowFile, delimiter=',', names=True, usecols=cols)
-            return data
-        except:
-            return None
+        fileext = os.path.splitext(LaharFlowFile)[1]
+
+        if fileext=='.txt':
+            try:
+                cols = csv_active_cols(LaharFlowFile)
+                data = np.genfromtxt(LaharFlowFile, delimiter=',', names=True, usecols=cols)
+                return data
+            except:
+                raise RuntimeWarning('Unable to get data from {}'.format(filename))
+        elif fileext=='.nc':
+            print ('Make nc work')
+
+
     
-    def raster(self, filename, var=None, full_domain=False, nodata=np.nan, masked=True, crs=None):
+    def raster(self, filename, var=None, full_domain=False, nodata=np.nan, masked=True, crs=None, res=None):
         filetype = self._file_type(filename)
         
         filename = self._valid_input_file(filename)
@@ -509,62 +619,181 @@ class LaharFlowData(object):
         else:
             data_cols = [x for x in var]
         
-        x = data['x_distance']
-        y = data['y_distance']
-
-        ux = np.unique(x)
-        uy = np.unique(y)
-        Nux = len(ux)
-        Nuy = len(uy)
-        if full_domain:
-            x0, y0= self.get_XY()
-            array = nodata*np.ones((self.nY,self.nX,len(data_cols)))
+        if res is None:
+            res_lat, res_lon = self._get_latlon_res()
         else:
-            x0 = ux
-            y0 = uy
-            array = nodata*np.ones((Nuy,Nux,len(data_cols)))
+            if isinstance(res, float):
+                res_lat = res
+                res_lon = res
+            elif isinstance(res, tuple) or isinstance(res, list):
+                res_lat = res[0]
+                res_lon = res[1]
+            else:
+                raise ValueError('res must be either a float or a tuple/list')
 
-        for k in range(0,len(x)):
-            xk = x[k]
-            yk = y[k]
-            ii = np.where(yk==y0)
-            jj = np.where(xk==x0)
-            for d, col in enumerate(data_cols):
-                array[ii,jj,d] = data[col][k]
-        
-        if masked:
-            array = np.ma.masked_where(array==nodata, array)
-        
+        tiles = data['tile'].astype(int)
+        uTiles = np.unique(tiles)
+
+
+        lats = data['latitude']
+        lons = data['longitude']
+
+        lat_pts = np.arange(lats.min(), lats.max()+res_lat, step=res_lat)
+        lon_pts = np.arange(lons.min(), lons.max()+res_lon, step=res_lon)
+        grid_lon, grid_lat = np.meshgrid(lon_pts, lat_pts)
         ds = xa.Dataset()
-        ds.coords["x"] = x0 + self.utm[0]
-        ds.coords["y"] = y0 + self.utm[1]
+        ds.coords["latitude"] = lat_pts
+        ds.coords["longitude"] = lon_pts
+        xy = np.zeros([len(lats),2])
+        xy[:,0] = lats
+        xy[:,1] = lons
+        uv = np.zeros([grid_lon.shape[0]*grid_lon.shape[1],2])
+        uv[:,0] = grid_lat.flatten()
+        uv[:,1] = grid_lon.flatten()
+        vtx, wts = interp_weights(xy, uv)
+        
         for d, col in enumerate(data_cols):
-            ds[col] = (("y","x"), array[:,:,d])
-            ds[col].rio.write_nodata(nodata, inplace=True)
+            vals = interpolate(data[col], vtx, wts)
+            grid_data = vals.reshape(grid_lat.shape[0], grid_lat.shape[1])
+         # grid_data = griddata((lats,lons), dat[col], (grid_lat, grid_lon), method='linear')
+            ds[col] = (("latitude","longitude"), grid_data)
             fmt = _get_LF_output_fmt(col, filetype=filetype)
             ds[col].attrs["short_name"] = fmt['short_name']
             ds[col].attrs["long_name"] = fmt['long_name']
             ds[col].attrs["units"] = fmt['units']
             ds[col].attrs["symbol"] = fmt['symbol']
+            ds[col].rio.write_nodata(np.nan, inplace=True)
+        ds['latitude'].attrs["units"] = "degrees"
+        ds['longitude'].attrs["units"] = "degrees"
+        ds.rio.write_crs(wgs84, inplace=True)
+
+        raster = ds
+
+        return raster
+
+        # tile_groups = self._connected_tiles(uTiles)
+
         
-        if filetype=='snapshot' and 'concentration' not in data_cols:
-            ds['concentration'] = ds['mass_of_solids']
-            ds['concentration'] = np.divide(ds['mass_of_solids'], ds['mass_per_unit_area'], where=ds['mass_per_unit_area'].where(ds['mass_per_unit_area'].data>1e-8))
-            fmt = _get_LF_output_fmt('concentration')
-            ds['concentration'].attrs["short_name"] = fmt['short_name']
-            ds['concentration'].attrs["long_name"] = fmt['long_name']
-            ds['concentration'].attrs["units"] = fmt['units']
-            ds['concentration'].attrs["symbol"] = fmt['symbol']
 
-        ds.x.attrs["units"] = "metres"
-        ds.y.attrs["units"] = "metres"
+        # raster = [xa.Dataset()]*len(tile_groups)
+
+        # for i, g in enumerate(tile_groups):
+
+        #     ii = np.isin(tiles, g)
+
+        #     dat = data[ii]
+
+        #     lats = dat['latitude']
+        #     lons = dat['longitude']
+
+        #     lat_pts = np.arange(lats.min(), lats.max()+res_lat, step=res_lat)
+        #     lon_pts = np.arange(lons.min(), lons.max()+res_lon, step=res_lon)
+
+        #     grid_lon, grid_lat = np.meshgrid(lon_pts, lat_pts)
+
+        #     ds = xa.Dataset()
+        #     ds.coords["latitude"] = lat_pts
+        #     ds.coords["longitude"] = lon_pts
+
+        #     xy = np.zeros([len(lats),2])
+        #     xy[:,0] = lats
+        #     xy[:,1] = lons
+
+        #     uv = np.zeros([grid_lon.shape[0]*grid_lon.shape[1],2])
+        #     uv[:,0] = grid_lat.flatten()
+        #     uv[:,1] = grid_lon.flatten()
+
+        #     vtx, wts = interp_weights(xy, uv)
+            
+        #     for d, col in enumerate(data_cols):
+        #         vals = interpolate(dat[col], vtx, wts)
+        #         grid_data = vals.reshape(grid_lat.shape[0], grid_lat.shape[1])
+
+        #         # grid_data = griddata((lats,lons), dat[col], (grid_lat, grid_lon), method='linear')
+        #         ds[col] = (("latitude","longitude"), grid_data)
+        #         fmt = _get_LF_output_fmt(col, filetype=filetype)
+        #         ds[col].attrs["short_name"] = fmt['short_name']
+        #         ds[col].attrs["long_name"] = fmt['long_name']
+        #         ds[col].attrs["units"] = fmt['units']
+        #         ds[col].attrs["symbol"] = fmt['symbol']
+        #         ds[col].rio.write_nodata(np.nan, inplace=True)
+        #     ds['latitude'].attrs["units"] = "degrees"
+        #     ds['longitude'].attrs["units"] = "degrees"
+        #     ds.rio.write_crs(wgs84, inplace=True)
+
+        #     raster[i] = ds
+
+        ## This could be useful, but doesn't work because of projection -- I think this needs to be fixed in laharflow
+        # for i, g in enumerate(tile_groups):
+
+        #     row_groups = self._tile_row_groups(g)
+            
+        #     for j, r in enumerate(row_groups):
+
+        #         for k, t in enumerate(r):
+        #             ds_tmp = xa.Dataset()
+        #             ii = np.where(tiles==t)
+            
+        #             dat = data[ii]
+        #             ds_tmp.coords["y"] = np.unique(dat['y_distance'])# + self.utm[1]
+        #             ds_tmp.coords["x"] = np.unique(dat['x_distance'])# + self.utm[0]
+        #             ds_tmp.coords['lat'] = (("y","x"), dat['latitude'].reshape(self.nYpertile, self.nXpertile))
+        #             ds_tmp.coords['lon'] = (("y","x"), dat['longitude'].reshape(self.nYpertile, self.nXpertile))
+        #             for d, col in enumerate(data_cols):
+        #                 ds_tmp[col] = (("y","x"), dat[col].reshape(self.nYpertile, self.nXpertile))
+        #                 fmt = _get_LF_output_fmt(col, filetype=filetype)
+        #                 ds_tmp[col].attrs["short_name"] = fmt['short_name']
+        #                 ds_tmp[col].attrs["long_name"] = fmt['long_name']
+        #                 ds_tmp[col].attrs["units"] = fmt['units']
+        #                 ds_tmp[col].attrs["symbol"] = fmt['symbol']
+        #             ds_tmp.coords["Northing"] = (("y"), ds_tmp.y.data + self.utm[1])
+        #             ds_tmp.coords["Easting"] = (("x"), ds_tmp.x.data + self.utm[0])
+                    
+            
+        #             if k==0:
+        #                 row = ds_tmp
+        #             else:
+        #                 row = xa.combine_nested([[row], [ds_tmp]], concat_dim=['x','y'])
+
+        #         if j==0:
+        #             raster[i] = row
+        #         else:     
+        #             raster[i] = xa.combine_nested([[raster[i], row]], concat_dim=['x','y'])
+            
+        #     raster[i].x.attrs["units"] = "metres"
+        #     raster[i].y.attrs["units"] = "metres"
         
-        ds = ds.rio.write_crs(self.utmCode)
+        #     raster[i] = raster[i].rio.write_crs(self.utmCode)
 
-        if crs is not None:
-            ds = ds.rio.reproject(crs)
 
-        return ds
+        # ds = xa.Dataset()
+        # ds.coords["x"] = x0-1.5*self.dX + self.utm[0] # This seems to be right, but why??
+        # ds.coords["y"] = y0-0.5*self.dY + self.utm[1]
+        # for d, col in enumerate(data_cols):
+        #     if np.isnan(nodata):
+        #         dA[d].where(np.isnan(dA[d])==False, drop=True)
+        #     else:
+        #         dA[d].where(dA[d]==nodata, drop=True)
+        #     ds[col] = dA[d]
+            
+        #     # ds[col] = (("y","x"), array[:,:,d])
+        #     ds[col].rio.write_nodata(nodata, inplace=True)
+        #     fmt = _get_LF_output_fmt(col, filetype=filetype)
+        #     ds[col].attrs["short_name"] = fmt['short_name']
+        #     ds[col].attrs["long_name"] = fmt['long_name']
+        #     ds[col].attrs["units"] = fmt['units']
+        #     ds[col].attrs["symbol"] = fmt['symbol']
+        
+        # if filetype=='snapshot' and 'concentration' not in data_cols:
+        #     ds['concentration'] = ds['mass_of_solids']
+        #     ds['concentration'] = np.divide(ds['mass_of_solids'], ds['mass_per_unit_area'], where=ds['mass_per_unit_area'].where(ds['mass_per_unit_area'].data>1e-8))
+        #     fmt = _get_LF_output_fmt('concentration')
+        #     ds['concentration'].attrs["short_name"] = fmt['short_name']
+        #     ds['concentration'].attrs["long_name"] = fmt['long_name']
+        #     ds['concentration'].attrs["units"] = fmt['units']
+        #     ds['concentration'].attrs["symbol"] = fmt['symbol']
+
+        return raster
 
     def _raster_flatten(raster):
         rflat = raster.data.flatten()
@@ -651,16 +880,28 @@ class LaharFlowData(object):
 
     def _get_max_depth(self):
         
-        _, _, h = self._get_data_column('MaxHeights.txt', 5)
-        
-        self.MaxDepth = np.amax(h)
+        if 'MaxHeights.txt' in self.txt_files:
+            _, _, h = self._get_data_column('MaxHeights.txt', 5)
+            self.MaxDepth = np.amax(h)
+            return
+        elif 'Maximums.nc' in self.nc_files:
+            # TODO: Make this work
+            return
+        else:
+            raise RuntimeWarning('No data to determine maximum depth')
     
     def max_depth(self, data=None, vmin=None):
 
         if data is None:
-            data = self.raster('MaxHeights.txt')
+            if 'MaxHeights.txt' in self.txt_files:
+                data = self.raster('MaxHeights.txt')
+            elif 'Maximums.nc' in self.nc_files:
+                # TODO: Make this work
+                print('problem')
+                return
+        
         if 'maximum_depth' not in list(data.keys()):
-            data = self.raster('MaxHeights.txt')
+            raise RuntimeError('maximum_depth not in data')
         
         maxh = data['maximum_depth']
 
@@ -671,16 +912,29 @@ class LaharFlowData(object):
 
     def _get_max_speed(self):
         
-        _, _, spd = self._get_data_column('MaxSpeeds.txt', 5)
-        
-        self.MaxSpeed = np.amax(spd)
+        if 'MaxSpeeds.txt' in self.txt_files:
+            _, _, spd = self._get_data_column('MaxSpeeds.txt', 5)
+            self.MaxSpeed = np.amax(spd)
+            return
+        elif 'Maximums.nc' in self.nc_files:
+            # TODO: Make this work
+            return
+        else:
+            raise RuntimeWarning('No data to determine maximum speed')
         
     def max_speed(self, data=None, vmin=None):
 
         if data is None:
-            data = self.raster('MaxSpeeds.txt')
+            if 'MaxSpeeds.txt' in self.txt_files:
+                data = self.raster('MaxSpeeds.txt')
+            elif 'Maximums.nc' in self.nc_files:
+                # TODO: Make this work
+                print('problem')
+                return
+        
         if 'maximum_speed' not in list(data.keys()):
-            data = self.raster('MaxSpeeds.txt')
+            raise RuntimeError('maximum_depth not in data')
+
         maxspd = data['maximum_speed']
 
         if vmin is not None:
@@ -939,16 +1193,22 @@ class LaharFlowData(object):
 
         fileIn = self._valid_input_file(fileIn)
 
-        raster = self.raster(fileIn, crs=crs)
+        raster = self.raster(fileIn)
+
+        if crs is not None:
+            if crs != wgs84:
+                for j, r in enumerate(raster):
+                    r.rio.reproject(crs, inplace=True)
 
         if fileOut is None:
-            fname = fileIn.replace('.txt', '.tif')
-            fileOut = os.path.join(self.dir, fname)
+            fname, fext = os.path.splitext(fileIn)
+            fileOut = os.path.join(self.dir, fname+'.tif')
         else:
             fileOut = os.path.join(self.dir, fileOut)
 
         print('Writing file to {}'.format(fileOut))
-        raster.rio.to_raster(fileOut, **kwargs)
+        raster.rio.to_raster(fileOut)
+        
         return
     
     def plot(self, result, variable,
